<?php

namespace Drupal\state_log;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\State\StateInterface;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;

/**
 * Logs changes to state.
 */
class StateLog implements StateInterface {

  use LoggerChannelTrait;

  /**
   * The key value store to use.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueStoreInterface
   */
  protected $keyValueStore;

  /**
   * The Config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected $configFactory;

  /**
   * Static state cache.
   *
   * @var array
   */
  protected $cache = [];

  /**
   * Constructs a new StateLog.
   *
   * @param \Drupal\Core\KeyValueStore\KeyValueFactoryInterface $key_value_factory
   *   The key value store to use.
   * @param \Drupal\Core\Config\ConfigFactory $config_factory
   *   The config factory.
   */
  public function __construct(KeyValueFactoryInterface $key_value_factory, ConfigFactory $config_factory) {
    $this->keyValueStore = $key_value_factory->get('state');
    $this->configFactory = $config_factory;
  }

  /**
   * {@inheritdoc}
   */
  public function get($key, $default = NULL) {
    $values = $this->getMultiple([$key]);
    return $values[$key] ?? $default;
  }

  /**
   * {@inheritdoc}
   */
  public function getMultiple(array $keys) {
    $values = [];
    $load = [];
    foreach ($keys as $key) {
      // Check if we have a value in the cache.
      if (isset($this->cache[$key])) {
        $values[$key] = $this->cache[$key];
      }
      // Load the value if we don't have an explicit NULL value.
      elseif (!array_key_exists($key, $this->cache)) {
        $load[] = $key;
      }
    }

    if ($load) {
      $loaded_values = $this->keyValueStore->getMultiple($load);
      foreach ($load as $key) {
        // If we find a value, even one that is NULL, add it to the cache and
        // return it.
        if (isset($loaded_values[$key]) || array_key_exists($key, $loaded_values)) {
          $values[$key] = $loaded_values[$key];
          $this->cache[$key] = $loaded_values[$key];
        }
        else {
          $this->cache[$key] = NULL;
        }
      }
    }

    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function set($key, $value) {
    $this->log($key, $value);
    $this->cache[$key] = $value;
    $this->keyValueStore->set($key, $value);
  }

  /**
   * {@inheritdoc}
   */
  public function setMultiple(array $data) {
    foreach ($data as $key => $value) {
      $this->log($key, $value);
    }
    foreach ($data as $key => $value) {
      $this->cache[$key] = $value;
    }
    $this->keyValueStore->setMultiple($data);
  }

  /**
   * {@inheritdoc}
   */
  public function delete($key) {
    $this->log($key, NULL);
    $this->deleteMultiple([$key]);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteMultiple(array $keys) {
    foreach ($keys as $key) {
      $this->log($key, NULL);
    }
    foreach ($keys as $key) {
      unset($this->cache[$key]);
    }
    $this->keyValueStore->deleteMultiple($keys);
  }

  /**
   * {@inheritdoc}
   */
  public function resetCache() {
    $this->cache = [];
  }

  /**
   * Returns whether the state_name is ignored.
   *
   * @param string $state_name
   *   The state name key.
   *
   * @return bool
   *   TRUE if the state key is ignored, FALSE otherwise.
   */
  private function isKeyIgnored(string $state_name): bool {
    $ignored_entities = $this->configFactory->get('state_log.settings')
      ->get('log_ignored_state') ?: [];
    foreach ($ignored_entities as $ignore) {
      if ($ignore == $state_name) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Logs the change to the state data.
   *
   * @param string $key
   *   The key of the data being logged.
   * @param mixed $value
   *   The data being logged.
   */
  protected function log(string $key, $value) {
    // Check and ignore configured keys.
    if (!$this->isKeyIgnored($key)) {
      // Only log items that have changed.
      if ($this->get($key) != $value) {
        // Log state changes.
        $this->getLogger('state_log')
          ->info('State changed: %key changed from %original to %value', [
            '%key' => $key,
            '%original' => var_export($this->get($key), TRUE),
            '%value' => var_export($value, TRUE),
          ]);
      }
    }
  }

}
